#include <stdio.h>
#include "BLE.h"
#include "esp_log.h"

#include "freertos/FreeRTOS.h"
#include "freertos/semphr.h"
#include "nimble/ble.h"
#include "esp_bt.h"
#include "esp_random.h"
#include "host/ble_hs.h"
#include "host/ble_uuid.h"
#include "host/util/util.h"
#include "nimble/nimble_port.h"
#include "nimble/nimble_port_freertos.h"
#include "services/gap/ble_svc_gap.h"
#include "services/gatt/ble_svc_gatt.h"

#define BLE_GATT_SERVICES_COUNT                     ((uint8_t) 1U)
#define BLE_GATT_OTA_CHARACTERISTICS_COUNT         ((uint8_t) 2U)

static const char * LogTitle = "BLE";

static uint8_t bleOwnAddressType = BLE_OWN_ADDR_PUBLIC;


static struct ble_gatt_svc_def gattServerServices[BLE_GATT_SERVICES_COUNT + 1];
static struct ble_gatt_chr_def gattOTAServiceCharacteristics[BLE_GATT_OTA_CHARACTERISTICS_COUNT + 1];


static const ble_uuid128_t GATT_OTAServiceUUID =
    BLE_UUID128_INIT(0x68, 0xa9, 0xce, 0xea, 0x18, 0xdf, 0x83, 0xa2, 0x02, 0x46,
                     0x03, 0x00, 0x00, 0x00, 0x00, 0x00);
                     
static const ble_uuid128_t GATT_OTADataCharacteristicUUID =
    BLE_UUID128_INIT(0x68, 0xa9, 0xce, 0xea, 0x18, 0xdf, 0x83, 0xa2, 0x02, 0x46,
                     0x03, 0x00, 0x02, 0x00, 0x00, 0x00);

static void hostTask(void *param);
static int gapEventHandler(struct ble_gap_event * event, void * arg);

static esp_err_t startAdvertising(void);
static int counter = 0;

int ReadData(struct os_mbuf *om, uint16_t min_len, uint16_t max_len, void *flat_buffer, uint16_t *written_length)
{
    int rc = 0;
    uint16_t bufferLength = 0;

    bufferLength = OS_MBUF_PKTLEN(om);
    if ((bufferLength < min_len) || (bufferLength > max_len)) {
        rc = BLE_ATT_ERR_INVALID_ATTR_VALUE_LEN;
    } else {
        rc = ble_hs_mbuf_to_flat(om, flat_buffer, max_len, written_length);
        if (rc != ESP_OK) {
            rc = BLE_ATT_ERR_UNLIKELY;
        }
    }

    return rc;
}

static int otaDataCharacteristicCallback(uint16_t conn_handle, uint16_t attr_handle, struct ble_gatt_access_ctxt * ctxt, void * arg)
{
    // Increment the counter for received packets
    static uint16_t dataSize = 0U;
    static uint8_t dataPacket[520];
    int rc = 0;
    uint16_t value;

    // Check if the GATT operation is a read or write
    if (ctxt->op == BLE_GATT_ACCESS_OP_WRITE_CHR) {
        rc = ReadData(ctxt->om, 1, 520, dataPacket, &dataSize );
        if (rc == 0)
        {
            value = (dataPacket[1] << 8) + dataPacket[0];
            ESP_LOGI(LogTitle, "otaDataCharacteristicCallback Received %d of size %d. Value %d", counter, dataSize, value); 
        }
        
    }
    counter++;

    return ESP_OK;
}

void InitializeGATTOtaServiceCharacteristics()
{
    gattOTAServiceCharacteristics[0].uuid = &GATT_OTADataCharacteristicUUID.u;
    gattOTAServiceCharacteristics[0].access_cb = otaDataCharacteristicCallback;
    gattOTAServiceCharacteristics[0].arg = NULL;
    gattOTAServiceCharacteristics[0].descriptors = NULL;
    gattOTAServiceCharacteristics[0].flags = BLE_GATT_CHR_F_WRITE_NO_RSP;
    gattOTAServiceCharacteristics[0].min_key_size = 0;
    gattOTAServiceCharacteristics[0].val_handle = NULL;

    gattOTAServiceCharacteristics[1].uuid = NULL;
}

void InitializeGattServices()
{
    // Data service
    gattServerServices[0].type = BLE_GATT_SVC_TYPE_PRIMARY;
    gattServerServices[0].uuid = &GATT_OTAServiceUUID.u;
    gattServerServices[0].includes = NULL;
    gattServerServices[0].characteristics = gattOTAServiceCharacteristics;

    gattServerServices[1].type = BLE_GATT_SVC_TYPE_END;
    InitializeGATTOtaServiceCharacteristics();

}

void InitializeGattServer()
{  
    int rc = ESP_OK;  
    ble_svc_gap_init();
    ble_svc_gatt_init();

    InitializeGattServices();


    rc = ble_gatts_count_cfg(gattServerServices);
    if(rc == ESP_OK)
    {
        // Add services
        rc = ble_gatts_add_svcs(gattServerServices);
        if(rc != ESP_OK)
        {
            ESP_LOGE(LogTitle, "ble_gatts_add_svcs error. rc=%d", rc);
        }
    }
    else
    {
        ESP_LOGE(LogTitle, "ble_gatts_count_cfg error. rc=%d", rc);
    }

}

static void bleHostResetCallback(int reason)
{
    ESP_LOGE(LogTitle, "BLE reset: reason = %d", reason);
}

static void bleHostSyncCallback(void)
{
    int rc;
    uint8_t addressValue[6] = {0};

    // Make sure we have address
    rc = ble_hs_util_ensure_addr(0);
    if(rc != 0)
    {
        ESP_LOGE(LogTitle, "Error ensuring address (%d)", rc);
        return;
    }

    // Figuring out address to use while advertising
    rc = ble_hs_id_infer_auto(0, &bleOwnAddressType);
    if(rc != 0)
    {
        ESP_LOGE(LogTitle, "Error determining address type (%d)", rc);
        return;
    }

    // Printing Address
    rc = ble_hs_id_copy_addr(bleOwnAddressType, addressValue, NULL);
    if(rc != 0)
    {
        ESP_LOGE(LogTitle, "Error copying address (%d)", rc);
        return;
    }

    ESP_LOGI(LogTitle, "Device Address: %02x:%02x:%02x:%02x:%02x:%02x",
             addressValue[5], addressValue[4], addressValue[3],
             addressValue[2], addressValue[1], addressValue[0]);

    // Start advertising
    rc = startAdvertising();
    if(rc != ESP_OK)
    {
        ESP_LOGE(LogTitle, "bleHostSyncCallback: startAdvertising error (%s)", esp_err_to_name(rc));
    }
}

static void gattServerRegisterCallback(struct ble_gatt_register_ctxt * ctxt, void * arg)
{
    char buf[BLE_UUID_STR_LEN];
    char buf2[BLE_UUID_STR_LEN];

    switch (ctxt->op)
    {
        case BLE_GATT_REGISTER_OP_SVC:
            ESP_LOGI(LogTitle, 
                "gattServerRegisterCallback: Registering service %s at address %p, ATT handle: 0x%04x.", 
                ble_uuid_to_str(ctxt->svc.svc_def->uuid, buf), 
                (void*)ctxt->svc.svc_def, 
                ctxt->svc.handle);
            break;

        case BLE_GATT_REGISTER_OP_CHR:
            ESP_LOGI(LogTitle, 
                "gattServerRegisterCallback: Registering characteristic %s in service %s  at address %p, ATT def handle: 0x%04x, ATT val handle: 0x%04x.", 
                ble_uuid_to_str(ctxt->chr.chr_def->uuid, buf), 
                ble_uuid_to_str(ctxt->chr.svc_def->uuid, buf2), 
                (void*)ctxt->chr.chr_def, 
                ctxt->chr.def_handle, 
                ctxt->chr.val_handle);
            break;

        case BLE_GATT_REGISTER_OP_DSC:
            ESP_LOGI(LogTitle, 
                "gattServerRegisterCallback: Registering descriptor %s at address %p, ATT handle: 0x%04x.", 
                ble_uuid_to_str(ctxt->dsc.dsc_def->uuid, buf), 
                (void*)ctxt->dsc.dsc_def, 
                ctxt->dsc.handle);
            break;

        default:
            ESP_LOGW(LogTitle, 
                "gattServerRegisterCallback: Unknown operation %d encountered. Context address: %p", 
                ctxt->op, 
                (void*)ctxt);
            break;
    }
}


void BLE_Initialize()
{
    ESP_LOGI(LogTitle, "BLE_Initialize");
    
    // Initialize the host stack.
    nimble_port_init();

    ble_hs_cfg.reset_cb = bleHostResetCallback;
    ble_hs_cfg.sync_cb = bleHostSyncCallback;
    ble_hs_cfg.gatts_register_cb = gattServerRegisterCallback;
    ble_hs_cfg.store_status_cb = ble_store_util_status_rr;
    ble_hs_cfg.sm_io_cap = BLE_HS_IO_NO_INPUT_OUTPUT;
    ble_hs_cfg.sm_sc = 0;

    InitializeGattServer();

    const char* name = "MYDEVICE";
    ble_svc_gap_device_name_set(name);
    
    nimble_port_freertos_init(hostTask);
    
}

static void hostTask(void *param)
{
    nimble_port_run();
    nimble_port_freertos_deinit();
}


static esp_err_t startAdvertising(void)
{
    counter = 0;
    static struct ble_gap_adv_params adv_params;
    struct ble_hs_adv_fields fields;
    int rc = ESP_OK;
    static char* advertisementName = "MYDEVICE";

    // Stop advertising if needed
    if(ble_gap_adv_active() == 0)
    {
        ble_gap_adv_stop();
    }

    memset(&fields, 0, sizeof fields);

    // Flags: discoverability + BLE only
    fields.flags = BLE_HS_ADV_F_DISC_GEN | BLE_HS_ADV_F_BREDR_UNSUP;

    // Setting power levels
    rc = esp_ble_tx_power_set(ESP_BLE_PWR_TYPE_ADV, ESP_PWR_LVL_P3);
    if(rc != ESP_OK)
    {
        ESP_LOGE(LogTitle, "esp_ble_tx_power_set(ESP_BLE_PWR_TYPE_ADV, ESP_PWR_LVL_N0) error (%s)", esp_err_to_name(rc));
        return rc;
    }

    rc = esp_ble_tx_power_set(ESP_BLE_PWR_TYPE_SCAN, ESP_PWR_LVL_N0);
    if(rc != ESP_OK)
    {
        ESP_LOGE(LogTitle, "esp_ble_tx_power_set(ESP_BLE_PWR_TYPE_SCAN, ESP_PWR_LVL_N0) error (%s)", esp_err_to_name(rc));
        return rc;
    }

    rc = esp_ble_tx_power_set(ESP_BLE_PWR_TYPE_CONN_HDL0, ESP_PWR_LVL_P3);
    if(rc != ESP_OK)
    {
        ESP_LOGE(LogTitle, "esp_ble_tx_power_set(ESP_BLE_PWR_TYPE_CONN_HDL0, ESP_PWR_LVL_P3) error (%s)", esp_err_to_name(rc));
        return rc;
    }

    rc = esp_ble_tx_power_set(ESP_BLE_PWR_TYPE_DEFAULT, ESP_PWR_LVL_P3);
    if(rc != ESP_OK)
    {
        ESP_LOGE(LogTitle, "esp_ble_tx_power_set(ESP_BLE_PWR_TYPE_DEFAULT, ESP_PWR_LVL_P3) error (%s)", esp_err_to_name(rc));
        return rc;
    }

    fields.name = (uint8_t*) advertisementName;
    fields.name_len = 8;
    fields.name_is_complete = 1;
    rc = ble_gap_adv_set_fields(&fields);
    if(rc == ESP_OK)
    {
        // Configure advertising
        memset(&adv_params, 0, sizeof adv_params);
        adv_params.conn_mode = BLE_GAP_CONN_MODE_UND; // undirected-connectable
        adv_params.disc_mode = BLE_GAP_DISC_MODE_GEN; // general-discoverable
        adv_params.itvl_min = 1000;                   // min interval in ms
        adv_params.itvl_max = 1500;                   // max interval in ms

        // Start advertising
        rc = ble_gap_adv_start(bleOwnAddressType, NULL, 10000, &adv_params, gapEventHandler, NULL);
        if(rc != ESP_OK)
        {
            ESP_LOGE(LogTitle, "Error enabling advertisement (%d)", rc);
        }
    }
    else
    {
        ESP_LOGE(LogTitle, "Error setting advertisement data (%d)", rc);
    }
    
    return rc;
}


static int gapEventHandler(struct ble_gap_event * event, void * arg)
{
    ESP_LOGW(LogTitle, "gapEventHandler BEGIN");
    int rc = 0;
    
    static struct ble_gap_upd_params connectionParams = {
        .itvl_min = 12, // Minimum value for connection interval in 1.25ms units
        .itvl_max = 12, // Maximum value for connection interval in 1.25ms units
        .latency = 0,   // Connection latency
        .supervision_timeout = 200, // Supervision timeout in 10ms units
        .min_ce_len = 12, // Minimum length of connection event in 0.625ms units
        .max_ce_len = 12, // Maximum length of connection event in 0.625ms units
    };

    switch (event->type)
    {        
        case BLE_GAP_EVENT_CONNECT:
            // A new connection was established or a connection attempt failed.

            rc = ble_gap_update_params(event->connect.conn_handle, &connectionParams);
            if(rc != 0)
            {
                ESP_LOGE(LogTitle, "Failed to update params; rc = %d", rc);
            }

            if(event->connect.status == 0)
            {
                ESP_LOGI(LogTitle, "Connection established. conn_handle = %d", event->connect.conn_handle);

                rc = ble_gap_conn_find(event->connect.conn_handle, NULL);

                // Stop advertising once a client is connected.
                rc = ble_gap_adv_stop();
                if( (rc != 0) && (rc != BLE_HS_EALREADY))
                {
                    ESP_LOGE(LogTitle, "ble_gap_adv_stop error (%d)", rc);
                    return rc;
                }

                // Increase power
                rc = esp_ble_tx_power_set(ESP_BLE_PWR_TYPE_CONN_HDL0, ESP_PWR_LVL_P3);
                if(rc != ESP_OK)
                {
                    ESP_LOGE(LogTitle, "esp_ble_tx_power_set error (%s)", esp_err_to_name(rc));
                    return rc;
                }

                if (rc != ESP_OK)
                {
                    ESP_LOGE(LogTitle, "gapEventHandler Fail to start connection timeout");
                    return rc;
                }
            } 
            else if (event->connect.status != 0) 
            {
                rc = startAdvertising();
                if(rc != ESP_OK)
                {
                    ESP_LOGE(LogTitle, "startAdvertising error (%s)", esp_err_to_name(rc));
                    return rc;
                }
            }
            return 0;

        case BLE_GAP_EVENT_DISCONNECT:
            // Connection terminated
            ESP_LOGI(LogTitle, "Disconnection. Reason: 0x%04x", event->disconnect.reason);

            // Start advertising again
            rc = startAdvertising();
            if(rc != ESP_OK)
            {
                ESP_LOGE(LogTitle, "startAdvertising error (%s)", esp_err_to_name(rc));
                return rc;
            }

            return 0;

        case BLE_GAP_EVENT_ADV_COMPLETE:
            ESP_LOGI(LogTitle, "Advertise complete. Reason: %d", event->adv_complete.reason);
            rc = startAdvertising();
            if(rc != ESP_OK)
            {
                ESP_LOGE(LogTitle, "startAdvertising error (%s)", esp_err_to_name(rc));
                return rc;
            }
            return 0;

        case BLE_GAP_EVENT_CONN_UPDATE:
            // The central has updated the connection parameters
            ESP_LOGI(LogTitle, "Connection updated. Status: %d. conn_handle=%d", event->conn_update.status, event->conn_update.conn_handle);
            rc = ble_gap_conn_find(event->conn_update.conn_handle, NULL);
            if(rc != 0)
            {
                ESP_LOGE(LogTitle, "ble_gap_conn_find error (%d)", rc);
            }
            return rc;

        case BLE_GAP_EVENT_ENC_CHANGE:
            // Encryption has been enabled or disabled for this connection
            ESP_LOGI(LogTitle, "Encryption change event. Status: %d", event->enc_change.status);
            rc = ble_gap_conn_find(event->enc_change.conn_handle, NULL);
            if(rc != 0)
            {
                ESP_LOGE(LogTitle, "ble_gap_conn_find error (%d)", rc);
            }
            return rc;

        case BLE_GAP_EVENT_SUBSCRIBE: 
            ESP_LOGI(LogTitle, "Subscribe Event:");
            ESP_LOGI(LogTitle, "\tconn_handle: %d", event->subscribe.conn_handle);
            ESP_LOGI(LogTitle, "\tattr_handle: 0x%04x", event->subscribe.attr_handle);
            ESP_LOGI(LogTitle, "\treason: %d", event->subscribe.reason);
            ESP_LOGI(LogTitle, "\tcur_notify: %d", event->subscribe.cur_notify);
            ESP_LOGI(LogTitle, "\tcur_indicate: %d", event->subscribe.cur_indicate);
            return 0;

        case BLE_GAP_EVENT_MTU:
            ESP_LOGI(LogTitle, "MTU update. conn_handle=%d, mtu=%d", event->mtu.conn_handle, event->mtu.value);
            return 0;

        case BLE_GAP_EVENT_REPEAT_PAIRING:
            ESP_LOGI(LogTitle, "Repeat pairing. conn_handle=%d", event->repeat_pairing.conn_handle);
            return 0;

        case BLE_GAP_EVENT_PASSKEY_ACTION:
            ESP_LOGI(LogTitle, "PASSKEY_ACTION_EVENT started");
            return 0;
        default:
            ESP_LOGW(LogTitle, "gapEventHandler Unhnadled event type %d", event->type);
            return 0;
    }

    return rc;
}